#!/usr/bin/env python3
from __future__ import annotations
from typing import Callable, Optional
from io import BytesIO
from sheet_stream import TableDocuments, IterRows, ListItems
from organize_stream.utils import ocr, cs, sp, sheet
from organize_stream.type_utils import (
    NotifyTableExtract, TextProgress, DiskFile, KeyFiles, KeyWordsFileName,
)
from organize_stream.read import read_image, read_document, OcrImage, concat_tables
import pandas as pd


class DocumentTextExtract(NotifyTableExtract):
    """

    """

    def __init__(
        self,
        recognize_image: ocr.RecognizeImage = OcrImage(), *,
        dpi: int = 200,
        apply_threshold: bool = False,
        notify_observers: bool = True,
        func_read_image: Callable[[cs.ImageObject, Optional[ocr.RecognizeImage]], TableDocuments] = None,
    ):
        super().__init__()
        if func_read_image is None:
            self._func_read_image = read_image
        else:
            self._func_read_image = func_read_image

        self.__collection_tables: ListItems[TableDocuments] = ListItems()
        self.__collection_tables.set_list_type(TableDocuments)
        self.recognize_image: ocr.RecognizeImage = recognize_image
        self.apply_threshold: bool = apply_threshold
        self.notify_observers: bool = notify_observers
        self.dpi: int = dpi
        self.__count_idx: int = 0
        self.text_progress: TextProgress = TextProgress()
        self.text_progress.set_pbar(sp.ProgressBarAdapter())
        self._pbar = self.text_progress.get_pbar()

    @property
    def values(self) -> list[TableDocuments]:
        return self.__collection_tables

    @property
    def length(self) -> int:
        return self.__collection_tables.length

    @property
    def pbar(self) -> sp.ProgressBarAdapter:
        return self.text_progress.get_pbar()

    @pbar.setter
    def pbar(self, pbar: sp.ProgressBarAdapter) -> None:
        self.text_progress.set_pbar(pbar)

    @property
    def is_empty(self) -> bool:
        return len(self.__collection_tables) == 0

    def add_table(self, tb: TableDocuments) -> None:
        if not isinstance(tb, TableDocuments):
            print(f'DEBUG: Tabela inválida: {tb}')
            return
        if tb.length == 0:
            print(f'DEBUG: Tabela vazia: {tb}')
            return
        self.__collection_tables.append(tb)
        self.__count_idx += 1
        if self.notify_observers:
            self.send_notify(tb)

    def add_dir_pdf(self, dir_pdf: sp.Directory) -> None:
        """
        Iterar sobre os arquivos PDF de uma pasta, extrair a tabela/texto de
        cada documento com OCR e adicionar cada tabela a propriedade/lista desse objeto.
        """
        files: list[sp.File] = sp.InputFiles(dir_pdf).pdfs
        self.text_progress.total = len(files)
        self.text_progress.set_default_text('Extraindo texto de Documento')
        self.text_progress.start_pbar()
        for n, f in enumerate(files):
            self.add_document(f)
        self.text_progress.stop_pbar()

    def add_dir_image(self, dir_image: sp.Directory):
        files_images = sp.InputFiles(dir_image).images
        self.text_progress.total = len(files_images)
        self.text_progress.set_default_text('Extraindo texto de imagem')
        self.text_progress.start_pbar()
        for idx, f in enumerate(files_images):
            self.add_image(f)
        self.text_progress.stop_pbar()

    def add_image(self, image: cs.ImageObject | sp.File | bytes | BytesIO) -> None:
        if isinstance(image, cs.ImageObject):
            pass
        else:
            image = cs.ImageObject(image, lib_image=cs.LibImage.OPENCV)
        if self.apply_threshold:
            image.set_threshold_black()
        self.add_table(self._func_read_image(image, self.recognize_image))

    def add_document(self, document: cs.DocumentPdf | sp.File | bytes | BytesIO) -> None:
        if isinstance(document, cs.DocumentPdf):
            pass
        elif isinstance(document, bytes):
            document = cs.DocumentPdf.create_from_bytes(BytesIO(document))
        else:
            document = cs.DocumentPdf(document)
        # Threshold
        if self.apply_threshold:
            _stream = cs.PdfStream()
            _stream.add_document(document)
            _stream.thresold()
            document = _stream.to_document()
        _tb = read_document(document, self.recognize_image, dpi=self.dpi, func_read_image=self._func_read_image)
        self.add_table(_tb)

    def to_table(self) -> TableDocuments:
        if len(self.__collection_tables) == 0:
            return TableDocuments.create_void_dict()
        return concat_tables(self.__collection_tables)

    def to_data(self) -> pd.DataFrame:
        return self.to_table().to_data().astype('str')

    def to_excel(self, file: sp.File) -> None:
        try:
            self.to_data().to_excel(file.absolute(), index=False)
        except Exception as e:
            print(f'Error: {e}')

    def read_image(self, image: DiskFile | cs.ImageObject) -> TableDocuments:
        if isinstance(image, cs.ImageObject):
            pass
        else:
            image = cs.ImageObject(image)
        return self._func_read_image(image, self.recognize_image)

    def read_document(self, document: DiskFile | cs.DocumentPdf) -> TableDocuments:
        if isinstance(document, cs.DocumentPdf):
            pass
        elif isinstance(document, bytes):
            document = cs.DocumentPdf.create_from_bytes(BytesIO(document))
        else:
            document = cs.DocumentPdf(document)
        return read_document(
            document,
            self.recognize_image,
            dpi=self.dpi,
            pbar=self.pbar,
            func_read_image=self._func_read_image
        )





#!/usr/bin/env python3
from __future__ import annotations

import os.path
import tempfile
from typing import Callable, Union
from io import BytesIO
from typing import Union
from sheet_stream import TableDocuments, ColumnsTable, ListItems
from organize_stream.type_utils import (
    FilterText, FilterData, DigitalizedDocument, LibDigitalized,
    ObserverTableExtraction, KeyWordsFileName, DiskFile, DynamicFile,
    DestFilePath,
)
from organize_stream.find import (
    NameFinderInnerText, NameFinderInnerData, OriginFileName, DestFilePath
)
from organize_stream.utils import (sp, cs, sheet)
from organize_stream.read import create_tb_from_names
from organize_stream.text_extract import DocumentTextExtract
from organize_stream.cartas import CartaCalculo, GenericDocument, FichaEpi
from organize_stream.erros import InvalidTDigitalizedDocument, InvalidSrcFile
import shutil
import zipfile
import pandas as pd


FindItem = Union[str, list[str]]


def move_list_files(
        mv_items: dict[str, list[sp.File]], *,
        replace: bool = False
) -> None:
    total_file = len(mv_items['src'])
    for idx, file in enumerate(mv_items['src']):
        output_path: sp.File = mv_items['dest'][idx]
        if not file.exists():
            print(f'[PULANDO]: {idx + 1} Arquivo não encontrado {file.absolute()}')
        if output_path.exists():
            if not replace:
                _count = 0
                origin_name = output_path.name_absolute()
                origin_ext = output_path.extension()
                while output_path.exists():
                    _count += 1
                    new_name = f'{origin_name}-{_count}{origin_ext}'
                    output_path = sp.File(new_name)
                del origin_name
                del origin_ext
        print(f'Movendo: {idx + 1}/{total_file} {file.absolute()}')
        try:
            shutil.move(file.absolute(), output_path.absolute())
        except Exception as e:
            print(f'{e}')
        del output_path


def move_path_files(
        mv_items: dict[OriginFileName, DestFilePath], *,
        replace: bool = False
) -> None:
    for _k in mv_items:
        output_path = mv_items[_k]
        if not _k.exists():
            print(f'[PULANDO O ARQUIVO NÃO EXISTE]: {_k.basename()}')
        if not replace:
            _count = 0
            origin_name: str = output_path.name_absolute()
            origin_ext = output_path.extension()
            while output_path.exists():
                _count += 1
                new_output_name: str = f'{origin_name}-{_count}{origin_ext}'
                output_path = sp.File(new_output_name)
            del origin_name
            del origin_ext
        try:
            shutil.move(_k.absolute(), output_path.absolute())
        except Exception as e:
            print(e)


def save_key_word_filename(
            key_word_file: KeyWordsFileName,
            out_dir: sp.Directory
        ) -> tuple[DynamicFile, DestFilePath | None, bool]:
    """
    SALVAR/MOVER o arquivo nomeado no disco e retorna Tuple contendo a sequência:
        - arquivo original, arquivo salvo no disco e booleano(sucesso ou falha na operação).
    Se o terceiro elemento for False, significa que a operação falhou.

    Se o arquivo origem (SRC) for bytes/BytesIO() os dados serão salvos no diretório informado
    sem alterar os arquivos de origem, se a origem for File()/str os arquivos serão movidos para
    o novo diretório.

    @rtype: tuple[DynamicFile, DestFilePath | None, bool]
    @type key_word_file: KeyWordsFileName
    @type out_dir: sp.Directory

    :rtype: Tuple[DynamicFile, DestFilePath | None, bool]
    :param out_dir: Diretório onde o arquivo de saída será gravado.
    :param key_word_file: Objeto/dicionário com os dados dos arquivos de origem e destino.
    """
    if key_word_file.input_dynamic_file is None:
        raise InvalidSrcFile()

    if key_word_file.output_name_str is None:
        return key_word_file.input_dynamic_file, None, False
    if key_word_file.extension_file is None:
        return key_word_file.input_dynamic_file, None, False

    out_dir.mkdir()
    if key_word_file.input_dynamic_file.is_bytes or key_word_file.input_dynamic_file.is_bytes_io:
        # Salvar os bytes no disco.
        return key_word_file.save(out_dir)
    elif key_word_file.input_dynamic_file.is_file or key_word_file.input_dynamic_file.is_file_path:
        # Mover o arquivo no disco.
        return key_word_file.move(out_dir)
    return key_word_file.input_dynamic_file, None, False


class NameFileInnerTable(object):

    def __init__(
                self,
                extractor: DocumentTextExtract = DocumentTextExtract(),
                lib_digitalized: LibDigitalized = LibDigitalized.GENERIC,
                filters: FilterText = None,
                func_save_file: Callable[
                    [KeyWordsFileName, sp.Directory], tuple[DynamicFile, DestFilePath | None, bool]
                ] = None,
            ):
        super().__init__()
        if func_save_file is None:
            self.func_save_file = save_key_word_filename
        else:
            self.func_save_file = func_save_file
        self.lib_digitalized: LibDigitalized = lib_digitalized
        self.extractor: DocumentTextExtract = extractor
        self.extractor.apply_threshold = True
        self.filters = filters
        # Dicionário para gravar o status de exportação dos arquivos,
        # sendo que as chaves apontam para o arquivo de origem DynamicFile() e
        # os valores apontam para uma tupla, (DestFilePath, bool).
        self.__exported_files: dict[DynamicFile, tuple[str | None, bool]] = {}
        self.__list_key_files: ListItems[KeyWordsFileName] = ListItems()
        self.__list_key_files.set_list_type(KeyWordsFileName)
        self.__temp_dir: sp.Directory = sp.Directory(tempfile.mkdtemp())

    def clear(self):
        self.__exported_files.clear()
        self.__list_key_files.clear()

    def get_exported_files(self) -> dict[DynamicFile, tuple[str | None, bool]]:
        return self.__exported_files

    def get_list_key_files(self) -> ListItems[KeyWordsFileName]:
        return self.__list_key_files

    def read_image(self, file: DiskFile | cs.ImageObject) -> KeyWordsFileName:
        """
            Gera um KeyWordsFileName que pode ser exportado/salvo no disco posteriormente.
        """
        __kw: KeyWordsFileName
        __dynamic: DynamicFile
        if isinstance(file, cs.ImageObject):
            __dynamic = DynamicFile(file.to_bytes())
            __kw = self._get_keyword_name_from_table(
                self.extractor.read_image(file)
            )
        else:
            __dynamic = DynamicFile(file)
            if __dynamic.is_bytes:
                __kw = self._get_keyword_name_from_table(
                    self.extractor.read_image(cs.ImageObject.create_from_bytes(__dynamic.file))
                )
            elif __dynamic.is_file_path:
                __kw = self._get_keyword_name_from_table(
                    self.extractor.read_image(cs.ImageObject.create_from_file(__dynamic.file))
                )
            else:
                __kw = self._get_keyword_name_from_table(
                    self.extractor.read_image(cs.ImageObject(__dynamic.file))
                )
        if __kw.extension_file is None:
            __kw.extension_file = '.png'
        __kw.input_dynamic_file = __dynamic
        return __kw

    def read_document(self, file: DiskFile | cs.DocumentPdf) -> KeyWordsFileName:
        __dynamic: DynamicFile
        __kw: KeyWordsFileName
        if isinstance(file, cs.DocumentPdf):
            __dynamic = DynamicFile(file.to_bytes())
            __kw = self._get_keyword_name_from_table(self.extractor.read_document(file))
        else:
            __dynamic = DynamicFile(file)
            __tb = self.extractor.read_document(file)
            __kw = self._get_keyword_name_from_table(__tb)

        if __kw.extension_file is None:
            __kw.extension_file = '.pdf'
        __kw.input_dynamic_file = __dynamic
        return __kw

    def _get_keyword_name_from_table(self, tb: TableDocuments) -> KeyWordsFileName:
        """
        Recebe uma tabela e retorna um dicionário de chave/valor com os dados
        do arquivo de origem e destino, incluindo extensão de arquivo.
        """
        key_words = KeyWordsFileName()
        _doc: DigitalizedDocument
        if self.lib_digitalized == LibDigitalized.GENERIC:
            _doc = GenericDocument(tb, filters=self.filters)
        elif self.lib_digitalized == LibDigitalized.CARTA_CALCULO:
            _doc = CartaCalculo.create(tb)
        elif self.lib_digitalized == LibDigitalized.EPI:
            _doc = FichaEpi.create(tb)
        else:
            raise InvalidTDigitalizedDocument(f'{__class__.__name__} Documento inválido: {self.lib_digitalized}')
        # Proteger o objeto gerado contra valores de str padrão.
        filename_str = _doc.get_output_name_str()
        src_extension = _doc.extension_file
        if (filename_str is not None) and (filename_str != 'nan') and (filename_str != ''):
            key_words.output_name_str = filename_str
        if (src_extension is not None) and (src_extension != '') and (src_extension != 'nan'):
            key_words.extension_file = src_extension
        return key_words

    def _save_file_keyword(
                self, key_word_file: KeyWordsFileName, out_dir: sp.Directory
            ) -> tuple[DynamicFile, DestFilePath, bool]:
        """
        Recebe um objeto KeyWordsFileName e um diretório para salvar o arquivo de origem
        no destino padronizado, se os arquivos fonte forem File()/str serão movidos, se não
        serão salvos sem alterar os arquivos fonte.
        """
        _status: tuple[DynamicFile, DestFilePath, bool] = self.func_save_file(key_word_file, out_dir)
        self.__exported_files[_status[0]] = None if _status[1] is None else _status[1].absolute(), _status[2]
        return _status

    def rename_image(self, image: DiskFile | cs.ImageObject, output_dir: sp.Directory):
        """
        Extrai o texto de uma imagem e renomeia conforme o padrão do documento informado nesse objeto.
        """
        __kw_im: KeyWordsFileName = self.read_image(image)
        self._save_file_keyword(__kw_im, output_dir)

    def rename_document(
                self, document: DiskFile | cs.DocumentPdf, output_dir: sp.Directory
            ) -> None:
        """
        Extrai o texto de um PDF e renomeia conforme o padrão do documento informado nesse objeto.
        """
        __kw_pdf: KeyWordsFileName = self.read_document(document)
        self._save_file_keyword(__kw_pdf, output_dir)

    def add_image(self, image: DiskFile | cs.ImageObject):
        __k_img = self.read_image(image)
        self.__list_key_files.append(__k_img)

    def add_document(self, document: DiskFile | cs.DocumentPdf):
        __k_doc: KeyWordsFileName = self.read_document(document)
        self.__list_key_files.append(__k_doc)

    def export_new_files(self, output_dir: sp.Directory) -> None:
        for __k in self.__list_key_files:
            self._save_file_keyword(__k, output_dir)

    def export_log_actions(self) -> pd.DataFrame:
        __data: dict[str, list[str]] = {
            'STATUS': [],
            'ARQUIVO_ORIGEM': [],
            'NOVO_NOME': [],

        }
        __key: DynamicFile
        current: tuple[str | None, bool]
        for __key in self.__exported_files.keys():
            current = self.__exported_files[__key]
            if __key is not None:
                __data['ARQUIVO_ORIGEM'].append(os.path.basename(__key.id_file))
                __data['NOVO_NOME'].append(os.path.basename(current[0]))
                __data['STATUS'].append("FALHA" if not current[1] else "SUCESSO")
        __df = pd.DataFrame(__data)
        return __df

    def export_keys_to_zip(self) -> BytesIO | None:
        if self.__list_key_files.length == 0:
            return None

        zip_buffer = BytesIO()
        key_file: KeyWordsFileName
        with zipfile.ZipFile(zip_buffer, "w") as zipf:
            for key_file in self.__list_key_files:
                id_file: DynamicFile = key_file.input_dynamic_file
                if id_file is None:
                    continue
                if (key_file.extension_file is None) or (key_file.output_name_str is None):
                    self.__exported_files[id_file] = (None, False)
                    continue

                dest_file_name: str = f'{key_file.output_name_str}{key_file.extension_file}'
                zipf.writestr(dest_file_name, key_file.input_dynamic_file.get_bytes())
                self.__exported_files[id_file] = (dest_file_name, True)
        zip_buffer.seek(0)
        return zip_buffer


class ExtractName(ObserverTableExtraction):

    def __init__(self, output_dir: sp.Directory, *, filters: FilterText = None):
        super().__init__()
        self._count: int = 0
        self.output_dir: sp.Directory = output_dir
        self.pbar: sp.ProgressBarAdapter = sp.ProgressBarAdapter()
        self.max_char: int = 90
        self.upper_case: bool = True
        self.save_tables: bool = True
        self.filters: FilterText = filters
        self.extractor: DocumentTextExtract = DocumentTextExtract()
        self.extractor.apply_threshold = False
        self.extractor.add_observer(self)

    @property
    def output_dir_tables(self) -> sp.Directory:
        return self.output_dir.concat('Tabelas', create=True)

    def _show_error(self, txt: str):
        print()
        self.pbar.update_text(f'{__class__.__name__} {txt}')

    def add_table(self, tb: TableDocuments):
        pass

    def export_tables(self, tb: TableDocuments) -> None:
        if not self.save_tables:
            return
        origin_name = tb.get_column(ColumnsTable.FILE_NAME)[0]
        output_path = self.output_dir_tables.join_file(f'{origin_name}.xlsx')
        if isinstance(output_path, sp.File):
            #print(f'DEBUG: Exportando ... {output_path.basename()}')
            tb.to_data().to_excel(output_path.absolute(), index=False)

    def export_final_table(self):
        if not self.save_tables:
            return
        self.extractor.to_excel(self.output_dir_tables.join_file('data.xlsx'))

    def receive_notify(self, notify: TableDocuments) -> None:
        pass

    def move_digitalized_doc(self, tb: TableDocuments) -> None:
        pass


class ExtractNameInnerText(ExtractName):
    """
    Mover/Renomear arquivos de acordo com padrões de texto presentes
    nos documentos/imagens.

    O padrão de texto a ser filtrado deve ser criado no objeto FilterText(). Se desejar
    filtrar mais de uma ocorrência nos documentos/imagens, separe as ocorrências com um '|'

    """

    def __init__(
                self,
                output_dir: sp.Directory, *,
                lib_digitalized: LibDigitalized = LibDigitalized.GENERIC,
                filters: FilterText = None,
            ):
        super().__init__(output_dir, filters=filters)
        self.lib_digitalized: LibDigitalized = lib_digitalized
        self.name_finder: NameFinderInnerText = NameFinderInnerText(self.output_dir)

    def receive_notify(self, notify: TableDocuments) -> None:
        self._count += 1
        self.move_digitalized_doc(notify)
        self.export_tables(notify)

    def add_table(self, tb: TableDocuments):
        self.move_digitalized_doc(tb)
        self.export_tables(tb)

    def move_digitalized_doc(self, tb: TableDocuments) -> None:
        """
        Mover/Renomear arquivos de acordo com padrões de texto presentes
        nos documentos/imagens.
        """
        dg: DigitalizedDocument
        if self.lib_digitalized == LibDigitalized.GENERIC:
            if self.filters is None:
                print(f'DEBUG: {__class__.__name__} Falha ... o filtro está vazio.')
                return
            dg = GenericDocument(tb, filters=self.filters)
        elif self.lib_digitalized == LibDigitalized.CARTA_CALCULO:
            dg = CartaCalculo.create(tb)
        elif self.lib_digitalized == LibDigitalized.EPI:
            dg = FichaEpi.create(tb)
        else:
            raise InvalidTDigitalizedDocument()
        new_names: dict[OriginFileName, DestFilePath] = self.name_finder.get_new_name(dg)
        move_path_files(new_names, replace=False)


class ExtractNameInnerData(ExtractName):
    """
        Organizar os arquivos com base nos dados de uma tabela/DataFrame
    """

    def __init__(self, output_dir: sp.Directory, *, filters: FilterData = None):
        super().__init__(output_dir, filters=None)
        self.filter_data: FilterData = filters
        self.name_inner_data: NameFinderInnerData = NameFinderInnerData(self.output_dir, filters=self.filter_data)

    def receive_notify(self, notify: TableDocuments) -> None:
        self._count += 1
        self.move_digitalized_doc(notify)
        self.export_tables(notify)

    def add_table(self, tb: TableDocuments):
        self.move_digitalized_doc(tb)
        self.export_tables(tb)

    def move_digitalized_doc(self, tb: TableDocuments) -> None:
        mv_items = self.name_inner_data.get_new_name(
            GenericDocument(tb, filters=None)
        )
        move_path_files(mv_items, replace=False)

    def move_where_math_filename(self, files: list[sp.File]) -> None:
        """
            Mover arquivos conforme as ocorrências de texto encontradas na tabela/DataFrame df.
        o nome do novo arquivo será igual à ocorrência de texto da coluna 'col_find', podendo
        estender o nome com elementos de outras colunas, tais colunas podem ser informadas (opcionalmente)
        no parâmetro cols_in_name.
            Ex:
        Suponha que a tabela para renomear aquivos tenha a seguinte estrutura:

        A      B        C
        maça   Cidade 1 xxyyy
        banana Cidade 2 yyxxx
        mamão  Cidade 3 xyxyx

        Se passarmos os parâmetros col_find='A' e col_new_name='A' e o texto banana for
        encontrado no(s) documento, o novo nome do arquivo será banana. Caso incluir o parâmetro
        cols_in_name=['B'] o novo nome do arquivo será banana-Cidade 2 ou
        banana-Cidade 2-yyxxx (se incluir cols_in_name=['B', 'C']).

        """
        values: list[TableDocuments] = create_tb_from_names(files)
        for current_tb in values:
            mv_items = self.name_inner_data.get_new_name(
                GenericDocument(current_tb, filters=None)
            )
            move_path_files(mv_items, replace=False)





class TableFileEmptyError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class InvalidTDigitalizedDocument(Exception):

    def __init__(self, *args):
        super().__init__(*args)


class InvalidSrcFile(Exception):

    def __init__(self, *args):
        super().__init__(*args)




#!/usr/bin/env python3
from __future__ import annotations
from abc import ABC, abstractmethod
import pandas as pd
from organize_stream.type_utils import (
    DigitalizedDocument, OriginFileName, DestFilePath,
    FilterData,

)

from organize_stream.utils import (
    ArrayString, ListString, ListColumnBody, ColumnsTable,
    sp, fmt_str_file, HeadValues, HeadCell, sheet
)


def get_column_values(df: pd.DataFrame, col: str) -> ArrayString:
    try:
        _values = df[col].astype('str').values.tolist()
    except Exception as e:
        print(e)
        return ArrayString([])
    else:
        return ArrayString(_values)


class SearchableText(object):
    default_elements: sheet.TableDocuments = sheet.TableDocuments.create_void_dict()
    default_columns: HeadValues = HeadValues([HeadCell(x) for x in list(default_elements.keys())])

    def __init__(self):
        self.elements: sheet.TableDocuments = sheet.TableDocuments.create_void_dict()

    def __repr__(self):
        return f'SearchableText\nHead: {self.head}\nBody: {self.body}'

    def is_empty(self) -> bool:
        return len(self.elements[HeadCell(ColumnsTable.TEXT)]) == 0

    @property
    def head(self) -> HeadValues:
        return HeadValues([HeadCell(x) for x in list(self.elements.keys())])

    @property
    def body(self) -> list[ListColumnBody]:
        return [ListColumnBody(HeadCell(_k), self.elements[_k]) for _k in self.elements.keys()]

    @property
    def first(self) -> dict[str, str]:
        if self.is_empty():
            return {}
        cols: HeadValues = self.head
        _first = {}
        for col in cols:
            _first[col] = self.elements[col][0]
        return _first

    @property
    def last(self) -> dict[str, str]:
        if self.is_empty():
            return {}
        cols = self.head
        _last = {}
        for col in cols:
            _last[col] = self.elements[col][-1]
        return _last

    @property
    def length(self) -> int:
        return len(self.elements[HeadCell(ColumnsTable.TEXT)])

    @property
    def files(self) -> ListColumnBody:
        return self.elements[HeadCell(ColumnsTable.FILE_PATH)]

    def get_item(self, idx: int) -> dict[str, str]:
        cols: HeadValues = self.head
        try:
            _item = {}
            for col in cols:
                _item[col] = self.elements[col][idx]
            return _item
        except Exception as err:
            print(err)
            return {}

    def get_column(self, name: str) -> ListColumnBody:
        return self.elements[HeadCell(name)]

    def add_line(self, line: dict[str, str]) -> None:
        cols_in_line: HeadValues = HeadValues([HeadCell(x) for x in list(line.keys())])
        cols_in_searchable: HeadValues = self.head
        for col in cols_in_searchable:
            if cols_in_line.contains(col, case=True, iqual=True):
                self.elements[col].append(line[col])

    def clear(self) -> None:
        for _k in self.elements.keys():
            self.elements[_k].clear()

    def to_string(self) -> str:
        """
            Retorna o texto da coluna TEXT em formato de string
        ou 'nas' em caso de erro nas = Not a String
        """
        try:
            return ' '.join(self.elements[HeadCell(ColumnsTable.TEXT)])
        except Exception as e:
            print(e)
            return 'nan'

    def to_data_frame(self) -> pd.DataFrame:
        return pd.DataFrame.from_dict(self.elements)

    def to_file_json(self, file: sp.File):
        """Exporta os dados da busca para arquivo .JSON"""
        dt = sp.JsonConvert.from_dict(self.elements).to_json_data()
        dt.to_file(file)

    def to_file_excel(self, file: sp.File):
        """Exporta os dados da busca para arquivo .XLSX"""
        self.to_data_frame().to_excel(file.absolute(), index=False)

    @classmethod
    def create(cls, df: pd.DataFrame) -> SearchableText:
        cols: list[str] = df.columns.tolist()
        _values: list[ListColumnBody] = []
        for col in cols:
            _values.append(
                ListColumnBody(
                    col, ListString(df[col].astype('str').values.tolist())
                )
            )
        s = cls()
        s.elements = sheet.TableDocuments(_values)
        return s


class NameFinder(ABC):

    def __init__(self, output_dir: sp.Directory):
        self.output_dir = output_dir

    @abstractmethod
    def get_new_name(self, digitalized: DigitalizedDocument) -> dict[OriginFileName, DestFilePath]:
        pass


class NameFinderInnerText(NameFinder):

    def __init__(self, output_dir: sp.Directory):
        super().__init__(output_dir)

    def get_new_name(self, digitalized: DigitalizedDocument) -> dict[OriginFileName, DestFilePath]:
        src_file = digitalized.file_path_origin
        filename = digitalized.get_output_name_with_extension()
        if (src_file is None) or (filename is None):
            return {}
        if filename == '':
            return {}
        output_path = self.output_dir.join_file(filename)
        return {OriginFileName(src_file.absolute()): DestFilePath(output_path.absolute())}


class NameFinderInnerData(NameFinder):

    def __init__(self, output_dir: sp.Directory, *, filters: FilterData):
        super().__init__(output_dir)
        self.filter_data = filters

    def get_values(self, df: pd.DataFrame, col: str) -> ArrayString:
        return get_column_values(df, col)

    def get_include_names(self, idx: int) -> str | None:
        if len(self.filter_data.cols_in_name) == 0:
            return None

        values = ArrayString([])
        for col in self.filter_data.cols_in_name:
            try:
                current_text = self.filter_data.src_df[col].astype('str').values.tolist()[idx]
            except Exception as e:
                pass
            else:
                if (current_text == 'nan') or (current_text == 'NaN') \
                        or (current_text == 'NaD') or (current_text == 'NaT') \
                        or (current_text == '') or (current_text is None) or (current_text == 'None'):
                    continue
                values.append(current_text)
        if values.length == 0:
            return None
        new_name = ''
        for i in values:
            new_name = f'{new_name}-{i}'
        return new_name

    def get_new_name(self, digitalized: DigitalizedDocument) -> dict[OriginFileName, DestFilePath]:
        extension_file = digitalized.extension_file
        _origin_path = digitalized.file_path_origin
        if (extension_file is None) or (_origin_path is None):
            return {}

        # Lista de valores da coluna texto.
        list_values_find: ArrayString = self.get_values(self.filter_data.src_df, self.filter_data.col_find)

        # Lista de valores da coluna com novos nomes de arquivo.
        content_new_names: ArrayString = self.get_values(self.filter_data.src_df, self.filter_data.col_new_name)

        # Lista de valores com as linhas de texto do arquivo em formato list[str].
        lines_in_doc: ArrayString = ArrayString(digitalized.get_lines_keys())

        line_df: str
        idx_df: int
        output_name: str = None
        for idx_df, line_df in enumerate(list_values_find):
            if not lines_in_doc.contains(line_df, case=False, iqual=False):
                continue

            output_name = content_new_names[idx_df]
            if (output_name == 'nan') or (output_name is None):
                output_name = ''
            include_strings = self.get_include_names(idx_df)
            if include_strings is not None:
                output_name = f'{output_name}-{include_strings}'
            if output_name == '':
                continue
            output_name: str = fmt_str_file(output_name)
            break

        if output_name is None:
            return {}

        output_name = f'{output_name}{extension_file}'
        _dest_path = self.output_dir.join_file(output_name)

        if not isinstance(_dest_path, sp.File):
            return {}
        if not isinstance(_origin_path, sp.File):
            return {}
        return {OriginFileName(_origin_path.absolute()): DestFilePath(_dest_path.absolute())}



from __future__ import annotations
from enum import StrEnum
from typing import TypeAlias, Union
from io import BytesIO
from organize_stream.utils import sp, sheet, ListString, ListColumnBody
from organize_stream.erros import InvalidSrcFile
from sheet_stream.type_utils import get_hash_from_bytes
import shutil

DiskFile: TypeAlias = Union[str, sp.File, bytes, BytesIO]


class LibDigitalized(StrEnum):

    GENERIC = 'generic'
    CARTA_CALCULO = 'carta_calculo'
    EPI = 'epi'


class KeyFiles(StrEnum):

    SRC_FILE_PATH = 'SRC_FILE_PATH'
    SRC_FILENAME = 'FILE_NAME'
    DIRECTORY = 'DIRECTORY'
    FILE_TYPE = 'FILE_TYPE'
    NEW_FILE_NAME = 'NEW_FILE_NAME'
    UNIQUE_KEY = 'UNIQUE_KEY'


class DynamicFile(object):

    def __init__(self, file: DiskFile):
        if isinstance(file, sp.File):
            self.__src_obj = 'FILE'
        elif isinstance(file, bytes):
            self.__src_obj = 'BYTES'
        elif isinstance(file, BytesIO):
            self.__src_obj = 'BYTES_IO'
        elif isinstance(file, str):
            self.__src_obj = 'STR'
        else:
            raise InvalidSrcFile(
                f'{__class__.__name__} Arquivo inválido ... {file}, use ... bytes|BytesIO|File|str'
            )
        self.file: DiskFile = file

    @property
    def id_file(self) -> str | None:
        if (self.__src_obj == 'BYTES') or (self.__src_obj == 'BYTES_IO'):
            _id_file = get_hash_from_bytes(self.file)
        elif self.__src_obj == 'STR':
            _id_file = self.file
        elif self.__src_obj == 'FILE':
            _id_file = self.file.absolute()
        else:
            _id_file = None
        return _id_file

    @property
    def is_bytes(self) -> bool:
        return isinstance(self.file, bytes)

    @property
    def is_bytes_io(self) -> bool:
        return isinstance(self.file, BytesIO)

    @property
    def is_file(self) -> bool:
        return isinstance(self.file, str)

    @property
    def is_file_path(self) -> bool:
        return isinstance(self.file, sp.File)

    def __hash__(self):
        return hash(self.file)

    def get_bytes(self) -> bytes:
        if self.is_bytes:
            return self.file
        elif self.is_bytes_io:
            self.file.seek(0)
            return self.file.getvalue()
        elif self.is_file:
            bt: bytes
            with open(self.file, 'rb') as f:
                bt = f.read()
            return bt
        elif self.is_file_path:
            bt: bytes
            with open(self.file.absolute(), 'rb') as f:
                bt = f.read()
            return bt
        return None


class OriginFileName(sp.File):

    def __init__(self, filename: str):
        super().__init__(filename)


class DestFilePath(sp.File):

    def __init__(self, filename: str):
        super().__init__(filename)


class KeyWordsFileName(dict):
    """
        Dicionário que contém informações de um arquivo no disco, bytes|BytesIO|File|str.
    A chave KeyFiles.NEW_FILE_NAME.value - pode ser definida futuramente para guardar o
    novo nome do arquivo (bastando concatenar com o diretório de saída para obter o caminho
    absoluto do novo arquivo).

    """

    def __init__(self):
        super().__init__({})
        self[KeyFiles.SRC_FILE_PATH.value] = None
        self[KeyFiles.SRC_FILENAME.value] = None
        self[KeyFiles.DIRECTORY.value] = None
        self[KeyFiles.FILE_TYPE.value] = None
        self[KeyFiles.NEW_FILE_NAME.value] = None
        self[KeyFiles.UNIQUE_KEY.value] = None

    @property
    def input_dynamic_file(self) -> DynamicFile | None:
        return self[KeyFiles.SRC_FILE_PATH.value]

    @input_dynamic_file.setter
    def input_dynamic_file(self, value: DynamicFile) -> None:
        if not isinstance(value, DynamicFile):
            raise TypeError(f'{__class__.__name__} Use: DynamicFile()')
        self[KeyFiles.SRC_FILE_PATH.value] = value

    @property
    def output_name_str(self) -> str | None:
        return self[KeyFiles.NEW_FILE_NAME.value]

    @output_name_str.setter
    def output_name_str(self, value: Union[str, None] | None) -> None:
        self[KeyFiles.NEW_FILE_NAME.value] = value

    @property
    def extension_file(self) -> str | None:
        return self[KeyFiles.FILE_TYPE]

    @extension_file.setter
    def extension_file(self, new: str):
        self[KeyFiles.FILE_TYPE] = new

    def __repr__(self):
        return f'KeyWordsFileNames: {super().__repr__()}'

    def __hash__(self):
        return hash(self.input_dynamic_file.file)

    def __eq__(self, other):
        if self.input_dynamic_file is not None and other.input_dynamic_file is not None:
            return self.input_dynamic_file.file == other.input_dynamic_file.file
        return self.__hash__() == other.__hash__()

    def keys(self) -> ListString:
        return ListString(list(super().keys()))

    def save(self, output_dir: sp.Directory) -> tuple[DynamicFile, DestFilePath | None, bool]:
        """
            Salva os bytes do arquivo original no novo caminho absoluto gerado.

        :param output_dir: Diretório para concatenar com o nome do novo arquivo
        (chave: KeyFiles.NEW_FILE_NAME.value).

        :return: Tuple (DynamicFile, DestFileName | None, bool).
        Se a operação falhar o terceiro elemento da tuple será False, se não, será True.
        tuple[0] -> DynamicFile arquivo original
        tuple[1] -> DestFileName caminho absoluto do arquivo salvo no disco ou None se a operação falhar.
        tuple[2] -> bool sucesso ou erro.

        """
        if self.extension_file is None:
            return self.input_dynamic_file, None, False
        if self.output_name_str is None:
            return self.input_dynamic_file, None, False

        try:
            output_file: sp.File = output_dir.join_file(f'{self.output_name_str}{self.extension_file}')
            output_dir.mkdir()
            with open(output_file.absolute(), 'wb') as f:
                f.write(self.input_dynamic_file.get_bytes())
        except Exception as e:
            print(f'{__class__.__name__} Error: {e}')
            return self.input_dynamic_file, None, False
        else:
            return self.input_dynamic_file, DestFilePath(output_file.absolute()), False

    def move(self, output_dir: sp.Directory) -> tuple[DynamicFile, DestFilePath | None, bool]:
        if (not self.input_dynamic_file.is_file) and (not self.input_dynamic_file.is_file_path):
            return self.input_dynamic_file, None, False
        if self.extension_file is None:
            return self.input_dynamic_file, None, False
        if self.output_name_str is None:
            return self.input_dynamic_file, None, False

        _count = 0
        output_file: sp.File = output_dir.join_file(f'{self.output_name_str}{self.extension_file}')
        if output_file.exists():
            while True:
                _count += 1
                output_file: sp.File = output_dir.join_file(f'{self.output_name_str}-{_count}{self.extension_file}')
                if not output_file.exists():
                    break
            
        try:
            if self.input_dynamic_file.is_file:
                shutil.move(self.input_dynamic_file.file, output_file.absolute())
            elif self.input_dynamic_file.is_file_path:
                shutil.move(self.input_dynamic_file.file.absolute(), output_file.absolute())
        except Exception as e:
            print(f'{__class__.__name__} Error: {e}')
            return self.input_dynamic_file, None, False
        else:
            return self.input_dynamic_file, DestFilePath(output_file.absolute()), True




from sheet_stream import clean_string, BAD_STRING_CHARS
from sheet_stream.type_utils import (
    HeadCell, HeadValues, ColumnsTable, ListColumnBody, ListString, ArrayString,
    TableDocuments, TableRow
)

import soup_files as sp
import convert_stream as cs
import ocr_stream as ocr
import sheet_stream as sheet

_remove_end_name: list[str] = ['-']
_remove_start_name: list[str] = ['-']


def remove_bad_chars(text: str) -> str:
    return clean_string(text)


def fmt_str_file(
            filename: str, *,
            max_char: int = 80,
            upper_case: bool = True
        ) -> str:
    filename = remove_bad_chars(filename)

    for c in _remove_end_name:
        if filename[-1] == c:
            filename = filename[:-1]
    for c in _remove_start_name:
        while c in filename[0]:
            filename = filename[1:]
    while '--' in filename:
        filename = filename.replace('--', '-')
    if upper_case:
        filename = filename.upper()
    if len(filename) <= max_char:
        return filename
    return filename[0:max_char]



#!/usr/bin/env python3
from __future__ import annotations
from abc import ABC, abstractmethod
from sheet_stream import TableDocuments


# Sujeito notificador
class NotifyProvider(ABC):

    def __init__(self):
        self._observers: list = []

    @property
    def observers(self) -> list:
        return self._observers

    @observers.setter
    def observers(self, value: list) -> None:
        pass

    @abstractmethod
    def add_observer(self, observer) -> None:
        pass

    @abstractmethod
    def send_notify(self, notify) -> None:
        pass


class NotifyTableExtract(NotifyProvider):

    def __init__(self):
        super().__init__()

    def add_observer(self, observer) -> None:
        if not isinstance(observer, ObserverTableExtraction):
            print(f'{__class__.__name__} Observador inválido')
            return
        self._observers.append(observer)

    def send_notify(self, notify) -> None:
        print(f'{__class__.__name__} Notificado!')
        observer: ObserverTableExtraction
        for observer in self._observers:
            observer.receive_notify(notify)


# Sujeito Observador.
class Observer(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def receive_notify(self, notify) -> None:
        pass


class ObserverTableExtraction(Observer):
    def __init__(self):
        super().__init__()

    def receive_notify(self, notify: TableDocuments) -> None:
        pass




#!/usr/bin/env python3
from __future__ import annotations
from typing import Callable, Optional
from organize_stream.utils import (
    cs, sp, sheet, ocr, HeadValues, ListColumnBody, HeadCell, ListString,
    ColumnsTable, TableDocuments,
)


class OcrImage(ocr.RecognizeImage):
    """
    Singleton para reconhecimento de texto em imagens
    """
    _instance = None  # armazena a instância única

    def __new__(cls):
        if cls._instance is None:
            # Cria a instância uma única vez
            cls._instance = super(OcrImage, cls).__new__(cls)
        return cls._instance

    def __init__(
                self, 
                bin_tess: ocr.BinTesseract = ocr.BinTesseract(), *,
                lib_ocr: ocr.LibOcr = ocr.LibOcr.PYTESSERACT,
            ):
        # Evita reexecutar __init__ em chamadas subsequentes
        if not hasattr(self, "_initialized"):
            super().__init__(bin_tess, lib_ocr=lib_ocr)
            self._initialized = True


def update_column_table(tb: TableDocuments, *, name: ColumnsTable, new_value: str) -> TableDocuments:
    col = tb.get_column(name)
    for idx, v in enumerate(col):
        col[idx] = new_value
    tb[name] = col
    return tb


def concat_tables(list_map: list[TableDocuments]) -> TableDocuments:
    if len(list_map) < 1:
        return TableDocuments.create_void_dict()
    _columns: HeadValues = list_map[0].columns
    list_values: list[ListColumnBody] = []
    text_table: TableDocuments
    col: ListColumnBody
    i: HeadCell

    for i in _columns:
        list_values.append(
            ListColumnBody(i, ListString([]))
        )
    for text_table in list_map:
        for col in list_values:
            col.extend(
                text_table[col.col_name]
            )
    final_tb = TableDocuments(list_values)
    col_idx = final_tb[sheet.ColumnsTable.KEY]
    for idx, v in enumerate(col_idx):
        col_idx[idx] = f'{idx}'
    final_tb[sheet.ColumnsTable.KEY] = col_idx
    return final_tb


def create_table_from_dict(data: dict[str, sheet.ListColumnBody]) -> sheet.TableDocuments:
    _values: list[sheet.ListColumnBody] = []
    for _k in data.keys():
        _values.append(data[_k])
    return sheet.TableDocuments(_values)


def create_tb_from_names(files: list[sp.File]) -> list[TableDocuments]:
    values: list[TableDocuments] = []
    for f in files:
        tb = TableDocuments.create_void_dict()
        tb[sheet.ColumnsTable.TEXT.value].append(f.name())
        tb[sheet.ColumnsTable.FILETYPE.value].append(f.extension())
        tb[sheet.ColumnsTable.FILE_PATH.value].append(f.absolute())
        tb[sheet.ColumnsTable.FILE_NAME.value].append(f.basename())
        tb[sheet.ColumnsTable.KEY.value].append('0')
        tb[sheet.ColumnsTable.NUM_PAGE.value].append('nan')
        tb[sheet.ColumnsTable.NUM_LINE.value].append('1')
        tb[sheet.ColumnsTable.DIR.value].append(f.dirname())
        values.append(tb)
    return values


def read_image(img: cs.ImageObject, recognize: ocr.RecognizeImage = OcrImage()) -> TableDocuments:
    txt_image = recognize.image_to_string(img)
    try:
        tb = TableDocuments.create_from_values(
            txt_image.split('\n'),
            file_path=img.metadata.file_path,
            dir_path=img.metadata.dir_path,
            file_type=img.metadata.extension,
        )
    except Exception as err:
        print('---------------------------------------------')
        print(f'DEBUG: falha ao tentar gerar a tabela de: {img.metadata.file_path}\n{err}')
        print('---------------------------------------------')
        return TableDocuments.create_void_dict()
    else:
        return tb


def read_document(
            document: cs.DocumentPdf,
            recognize: ocr.RecognizeImage = OcrImage(), *,
            pbar: sp.ProgressBarAdapter = sp.ProgressBarAdapter(),
            dpi: int = 200,
            func_read_image: Callable[[cs.ImageObject, Optional[ocr.RecognizeImage]], TableDocuments] = None,
        ) -> TableDocuments:
    """
    Aplicar OCR em documento PDF e retornar uma tabela
    dos textos presentes no documento.
    """
    if func_read_image is None:
        func_read_image = read_image
    list_tables: list[TableDocuments] = []
    text_progress = sp.TextProgress()
    text_progress.pbar = pbar
    text_progress.start_pbar()
    text_progress.pbar.update(0, 'Iniciando a extração da tabela PDF')

    convert = cs.ConvertPdfToImages.create(document)
    convert.set_pbar(pbar)
    images: list[cs.ImageObject] = convert.to_images(dpi=dpi)
    text_progress.total = len(images)
    text_progress.text = 'OCR PDF'

    for page_pdf_idx, img in enumerate(images):
        text_progress.set_update()
        current_tb: TableDocuments = func_read_image(img, recognize)
        if current_tb.length > 0:
            current_tb = update_column_table(
                current_tb, name=sheet.ColumnsTable.FILETYPE, new_value=document.metadata.extension
            )
            current_tb = update_column_table(
                current_tb, name=sheet.ColumnsTable.NUM_PAGE, new_value=f'{page_pdf_idx+1}'
            )
            list_tables.append(current_tb)
    text_progress.pbar.update(100, 'Extração finalizada!')
    text_progress.stop_pbar()
    return concat_tables(list_tables)






from __future__ import annotations
import soup_files as sp
from organize_stream import fmt_str_file
from organize_stream.type_utils import DigitalizedDocument, FilterText
from organize_stream.utils import remove_bad_chars
from sheet_stream import (
    ArrayString, ListColumnBody, TableDocuments,
    BAD_STRING_CHARS, ListString, ColumnsTable, ConvertStringDate, LibDate
)
import pandas as pd


class CartaCalculo(DigitalizedDocument):

    default_filter = FilterText(
            'TOI',
            separator=' ',
            key_words=['UC', 'TOI', 'POSTAGEM', 'LIVRO'],
            iqual=False,
            case=False,
        )

    def __init__(self, tb: TableDocuments, *, filters: FilterText):
        super().__init__(tb, filters=filters)

        self.localidades: dict[str, str] = {
            'NOVA MA': 'NOVA MAMORE',
            'MAMORE': 'NOVA MAMORE',
            'GUAJAR': 'GUAJARA MIRIM',
            'GUAJ': 'GUAJARA MIRIM',
            'VISTA': 'VISTA ALEGRE',
            'EXTREM': 'EXTREMA',
        }

    @classmethod
    def create(cls, tb: TableDocuments) -> CartaCalculo:
        return cls(tb, filters=cls.default_filter)

    @property
    def cidade(self) -> str | None:
        lines: ListColumnBody = self.lines
        _loc = None
        for k in self.localidades.keys():
            out = lines.find_text(k)
            if out is not None:
                _loc = self.localidades[k]
                break
        if _loc is None:
            if lines.contains('LOCA'):
                arr = ArrayString([])
                for line in lines:
                    if ' ' in line:
                        arr.extend(line.split(' '))
                    else:
                        arr.add_item(line)
                _loc = arr.get_next_all('LOCA')
                _loc = ' '.join(_loc)
                if len(_loc) > 10:
                    _loc = _loc[:10]

        if _loc is None:
            return None
        return remove_bad_chars(_loc).upper()

    @property
    def medidor(self) -> str | None:
        _medidor = self.lines.find_text('MEDI')

        if _medidor is None:
            return None
        arr = ArrayString(_medidor.split(' '))
        arr = arr.get_next_all('MEDI')
        if arr.is_empty:
            return None
        final_medidor = ' '.join(arr)
        if len(final_medidor) > 12:
            final_medidor = final_medidor[:12]
        return remove_bad_chars(final_medidor)

    def get_line_key(self) -> str | None:
        _check = ['UC', 'TOI', 'TOL']
        _key_word = 'CAR'
        lines = self.lines
        filter_list = ArrayString([])
        elements = ArrayString([])
        list_index: list[int] = []

        # Filtrar os indices desejados.
        for txt in _check:
            idx = lines.find_index(txt)
            if idx is not None:
                if not idx in list_index:
                    list_index.append(idx)

        # Gerar nova lista com os valores filtrados.
        for num in list_index:
            if ' ' in lines[num]:
                filter_list.extend(lines[num].split(' '))
            else:
                filter_list.append(lines[num])

        # Gerar a linha final
        for item in _check:
            i = filter_list.get_next_string(item)
            if i is not None:
                elements.append(i)

        if elements.length == 0:
            final_line: str = ''
        else:
            final_line: str = ' '.join(elements)

        # Incluir a cidade
        cidade = self.cidade
        medidor = self.medidor
        if medidor is not None:
            final_line = f'{final_line}-{medidor}'
        if cidade is not None:
            final_line = f'{final_line}-{cidade}'
        return remove_bad_chars(final_line)

    def get_lines_keys(self) -> ArrayString:
        lines = self.lines
        content: ArrayString = ArrayString([])
        for k in self.uniq_key_words:
            values = lines.find_text(k)
            if values is not None:
                elements = ArrayString(values.split(' '))
                uniq_value = elements.get_next_string(k)
                if uniq_value is not None:
                    content.append(f'{remove_bad_chars(uniq_value)}')
        content.append(self.cidade)
        content.append(self.medidor)
        return content

    def get_output_name_str(self) -> str | None:
        line_key = self.get_line_key()
        if (line_key is None) or (line_key == ''):
            return None
        line_key = fmt_str_file(line_key)
        return line_key


class GenericDocument(DigitalizedDocument):

    def __init__(self, tb: TableDocuments, *, filters: FilterText):
        super().__init__(tb, filters=filters)

    def _get_value_with_str(self, tb: TableDocuments) -> str | None:
        list_new_names: ListString = ListString([])
        tb_txt_file: pd.DataFrame = pd.DataFrame.from_dict(tb)
        df = tb_txt_file[[ColumnsTable.TEXT, ColumnsTable.FILE_PATH]].astype('str')

        # Divide padrões múltiplos separados por "|"
        patterns = [p.strip() for p in self.filters.find_txt.split('|') if p.strip()]
        if not patterns:
            print(f'{__class__.__name__}: Nenhum padrão de busca válido informado.')
            return None

        # Define padrão regex dependendo de "iqual"
        if self.filters.iqual:
            # AQUI estava: '^(' + '|'.join(patterns) + ')$'
            regex_pattern = '^(?:' + '|'.join(patterns) + ')$'  # Mude para (?:...)
        else:
            # AQUI estava: '(' + '|'.join(patterns) + ')'
            regex_pattern = '(?:' + '|'.join(patterns) + ')'  # Mude para (?:...)

        # Filtra linhas no DataFrame
        mask: pd.Series = df[ColumnsTable.TEXT].str.contains(
            regex_pattern,
            case=self.filters.case,
            regex=True,
            na=False
        )
        matched_df = df[mask]
        total_matches = len(matched_df)

        if total_matches == 0:
            return None

        # Para cada linha encontrada, gera nome limpo e adiciona à lista de movimentação
        for _, row in matched_df.iterrows():
            current_line: str = row[ColumnsTable.TEXT]
            if self.uniq_key_words.length > 0:
                for i in self.uniq_key_words:
                    if not i.upper() in current_line.upper():
                        continue
            # Usa o texto da linha como base do novo nome
            current_output_name: str = fmt_str_file(current_line.strip())
            if len(current_output_name) < 4:
                continue
            # Garante que o nome não fique vazio
            if not current_output_name:
                continue
            list_new_names.append(current_output_name)

        if list_new_names.length == 0:
            return None

        output_name: str = ' '.join(list_new_names)
        if len(output_name) < 4:
            return None
        return output_name

    def get_line_key(self) -> str:
        pass

    def get_output_name_str(self) -> str | None:
        if self.tb.get_column(ColumnsTable.TEXT).length == 0:
            return None
        _filename = self._get_value_with_str(self.tb)
        return _filename


class FichaEpi(GenericDocument):

    def __init__(self, tb: TableDocuments, *, filters: FilterText):
        super().__init__(tb, filters=filters)
        self.dias = [
            'segunda-feira', 'terca-feira', 'quarta-feira', 'quinta-feira',
            'sexta-feira', 'sabado',
        ]

    @classmethod
    def create(cls, tb: TableDocuments) -> FichaEpi:
        fil = FilterText('MATR', key_words=['NOME'])
        return cls(tb, filters=fil)

    def get_line_key(self) -> str:
        return self.get_nome()

    def get_output_name_with_extension(self) -> str | None:
        line_date = self.get_date_doc()
        filename = self.get_nome()
        _extension = self.extension_file
        if filename is None:
            return None
        if _extension is None:
            return None
        if line_date is not None:
            filename = f'{filename}-{line_date}'
        filename = fmt_str_file(filename)
        return f'{filename}{_extension}'

    def get_output_name_str(self) -> str | None:
        line_date = self.get_date_doc()
        filename = self.get_nome()

        if filename is None:
            return None
        if line_date is not None:
            filename = f'{filename}-{line_date}'
        return fmt_str_file(filename)

    def get_nome(self) -> str | None:
        return self.tb.get_column(ColumnsTable.TEXT).get_next_string('MATR')

    def get_date_doc(self) -> str | None:
        lines = self.lines
        _date_line: str = None
        _date_txt: str = None
        for d in self.dias:
            txt = lines.find_text(d)
            if txt is not None:
                _date_line = txt
                _date_txt = d
                break
        if _date_line is None:
            return None
        idx = _date_line.find(_date_txt)
        _date_line = _date_line[idx:].replace(_date_txt, '').strip()
        try:
            arr = ArrayString(_date_line.split(' '))
            conv: dict = ConvertStringDate().month_to_number
            _day = arr.get_back_string('de')
            _year = arr[-1]
            _month = arr.get_next_string('de').strip().lower()
            _month = conv[_month]
        except Exception as e:
            pass
        else:
            _date_line = f'{_year}-{_month}-{_day}'.replace('.', '')
        return _date_line



from __future__ import annotations
from abc import ABC, abstractmethod
import pandas as pd
from organize_stream.erros import TableFileEmptyError
from sheet_stream import (
    TableDocuments, ArrayString, ColumnsTable,
    ListColumnBody
)
import soup_files as sp


class FilterText(object):
    """
        Padrão de informações a serem filtradas em um documento.
    """

    def __init__(
                self,
                find_txt: str, *,
                separator: str = ' ',
                case: bool = False,
                iqual: bool = False,
                key_words: list[str] = None,
            ):
        self.find_txt: str = find_txt
        self.case: bool = case
        self.iqual: bool = iqual
        self.separator: str = separator
        self.key_words: list[str] = key_words


class FilterData(object):

    def __init__(
                self,
                src_df: pd.DataFrame, *,
                col_find: str,
                col_new_name: str,
                cols_in_name: list[str],
            ):
        self.col_find: str = col_find
        self.col_new_name: str = col_new_name
        self.cols_in_name: list[str] = cols_in_name
        self.src_df: pd.DataFrame = src_df.astype('str')


class DigitalizedDocument(ABC):

    default_filter: FilterText | None = None

    def __init__(self, tb: TableDocuments, *, filters: FilterText):
        self.tb: TableDocuments = tb
        if self.tb.length == 0:
            raise TableFileEmptyError('A tabela de arquivos não pode estar vazia!')
        self.filters: FilterText = filters

    @property
    def uniq_key_words(self) -> ArrayString:
        return ArrayString(self.filters.key_words)

    @property
    def file_path_origin(self) -> sp.File | None:
        value: ListColumnBody = self.tb.get_column(ColumnsTable.FILE_PATH)
        if value.is_empty:
            return None
        if (value[0] == '') or (value[0] == 'nan') or (value[0] == 'None') \
                or (value[0] == '-') or (value[0] == 'NaT'):
            return None
        try:
            file_path = sp.File(value[0])
        except Exception as e:
            print(e)
            return None
        else:
            if file_path.path.exists():
                return file_path
            return None

    @property
    def dir_path_origin(self) -> sp.Directory | None:
        value: ListColumnBody = self.tb.get_column(ColumnsTable.DIR)
        if value.is_empty:
            return None
        if (value[0] == '') or (value[0] == 'nan') or (value[0] == 'None') \
                or (value[0] == '-') or (value[0] == 'NaT'):
            return None
        try:
            _dir_path = sp.Directory(value[0])
        except Exception as e:
            print(e)
            return None
        else:
            if _dir_path.path.exists():
                return _dir_path
            return None

    @property
    def extension_file(self) -> str | None:
        value: ListColumnBody = self.tb.get_column(ColumnsTable.FILETYPE)
        if value.is_empty:
            return None
        if (value[0] == '') or (value[0] == 'nan') or (value[0] == 'None') \
                or (value[0] == '-') or (value[0] == 'NaT'):
            return None
        return value[0]

    @property
    def lines(self) -> ListColumnBody:
        return self.tb.get_column(ColumnsTable.TEXT)

    def __repr__(self):
        return f'{__class__.__name__}: {self.get_line_key()}'

    @abstractmethod
    def get_line_key(self) -> str:
        pass

    def get_lines_keys(self) -> ArrayString:
        return self.tb.get_column(ColumnsTable.TEXT)

    def get_output_name_with_extension(self) -> str | None:
        """Retorna o novo nome do arquivo, incluindo a extensão"""
        if (self.extension_file is None) or (self.extension_file == '') or (self.extension_file == 'nan'):
            return None
        if (self.get_output_name_str() is None) or (self.get_output_name_str() == '') or (self.get_output_name_str() == 'nan'):
            return None
        return f'{self.get_output_name_str()}{self.extension_file}'

    @abstractmethod
    def get_output_name_str(self) -> str | None:
        """
        Retorna o novo nome do arquivo, sem a extensão
        """
        pass

    def to_excel(self, file: sp.File):
        self.tb.to_data().to_excel(file.absolute())

    def to_file_text(self, file: sp.File):
        lines = self.lines
        print(f'Exportando: {file.absolute()}')
        with open(file.absolute(), 'w', encoding='utf-8') as f:
            f.writelines(lines)



from .observer import (
    Observer, NotifyProvider, ObserverTableExtraction, NotifyTableExtract
)
from .digital_doc import DigitalizedDocument, FilterText, FilterData
from .keyword_files import (
    DestFilePath, OriginFileName, LibDigitalized, KeyFiles,
    KeyWordsFileName, DiskFile, DynamicFile
)
from enum import StrEnum
from soup_files import File, ProgressBarAdapter


class TextProgress(object):

    def __init__(self, total: int = 1, start_value: int = 0):
        self.start_value = start_value
        self.total = total
        if total < start_value:
            raise ValueError(f'Total {total} is less than start value {start_value}')
        if total == 0:
            raise ValueError(f'Total {total} is zero')
        self.__default_text: str = 'Progresso'
        self.__pbar: ProgressBarAdapter = ProgressBarAdapter()

    def set_default_text(self, text: str):
        self.__default_text = text

    def set_pbar(self, pbar: ProgressBarAdapter):
        self.__pbar = pbar

    def get_pbar(self) -> ProgressBarAdapter:
        return self.__pbar

    def start_pbar(self):
        self.__pbar.start()

    def stop_pbar(self):
        self.__pbar.stop()

    def set_update(self, text: str = None):
        if text is None:
            self.__pbar.update(
                ((self.start_value+1) / self.total) * 100,
                self.__default_text,
            )
        else:
            self.__pbar.update(
                ((self.start_value + 1) / self.total) * 100,
                f'{text}',
            )
        self.start_value += 1


__all__ = [
    'DigitalizedDocument', 'FilterText', 'Observer',
    'NotifyProvider', 'LibDigitalized', 'FilterData',
    'OriginFileName', 'DestFilePath', 'TextProgress',
    'NotifyTableExtract', 'ObserverTableExtraction',
    'KeyFiles', 'KeyWordsFileName', 'DiskFile', 'DynamicFile',
]




from .find_names import (
    NameFinderInnerData, NameFinder, NameFinderInnerText, SearchableText,
    get_column_values, OriginFileName, DestFilePath
)

__all__ = [
    'NameFinder', 'NameFinderInnerText', 'NameFinderInnerData', 'SearchableText',
    'OriginFileName', 'DestFilePath'
]



#!/usr/bin/env python3

__version__ = '2.4.7'
from .utils import fmt_str_file, remove_bad_chars, clean_string, BAD_STRING_CHARS
from .type_utils import (
    DigitalizedDocument, FilterText, FilterData, LibDigitalized, Observer, NotifyProvider,
)
from .read import (
    read_image, read_document, create_tb_from_names
)
from .find import (
    SearchableText, NameFinderInnerText, NameFinderInnerData, NameFinder,
    OriginFileName, DestFilePath
)
from .text_extract import DocumentTextExtract
from .document import (
    ExtractNameInnerData, ExtractNameInnerText, NameFileInnerTable
)
from .cartas import CartaCalculo, GenericDocument






from .carta_calculo import (
    CartaCalculo, GenericDocument, DigitalizedDocument, FichaEpi
)
from organize_stream.utils import fmt_str_file
import soup_files as sp
import shutil


def move_cartas(cartas: list[DigitalizedDocument], output_dir: sp.Directory):
    output_dir.mkdir()
    for carta in cartas:
        try:
            src_file: sp.File = carta.file_path_origin
            if src_file is None:
                continue
            if not src_file.exists():
                continue
            output_file_name = fmt_str_file(carta.get_line_key())

            if output_file_name is None:
                continue
            if carta.extension_file is None:
                continue
            output_file_name = f'{output_file_name}{carta.extension_file}'
        except Exception as err:
            print(err)
        else:
            dest_file = output_dir.join_file(output_file_name)
            if not isinstance(dest_file, sp.File):
                continue
            if dest_file.exists():
                continue
            try:
                print(f'Movendo: {dest_file.absolute()}')
                shutil.move(src_file.absolute(), dest_file.absolute())
            except Exception as e:
                print(e)


__all__ = [
    'CartaCalculo', 'GenericDocument', 'move_cartas', 'FichaEpi'
]
